package djmil.cordacheckers.contracts;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import djmil.cordacheckers.states.GameProposalResolutionState;
import djmil.cordacheckers.states.GameProposalState;
import net.corda.v5.base.exceptions.CordaRuntimeException;
import net.corda.v5.ledger.utxo.Command;
import net.corda.v5.ledger.utxo.transaction.UtxoLedgerTransaction;

public class GameProposalContract implements net.corda.v5.ledger.utxo.Contract {

    private final static Logger log = LoggerFactory.getLogger(GameProposalContract.class);

    public static class Create implements Command { }
    public static class Accept implements Command { }
    public static class Reject implements Command { }
    public static class Cancel implements Command { }

    @Override
    public void verify(UtxoLedgerTransaction trx) {
        log.info("GameProposalContract.verify() called");

        requireThat(trx.getCommands().size() == 1, REQUIRE_SINGLE_COMMAND);
        Command command = trx.getCommands().get(0);

        if (command instanceof Create) {
            requireThat(trx.getInputContractStates().isEmpty(), CREATE_INPUT_STATE);
            requireThat(trx.getOutputContractStates().size() == 1, CREATE_OUTPUT_STATE);
            
            GameProposalState outputState = trx.getOutputStates(GameProposalState.class).get(0);
            requireThat(outputState != null, CREATE_OUTPUT_STATE);

            requireThat(outputState.getRecipientColor() != null, NON_NULL_RECIPIENT_COLOR);
        } else 
        if (command instanceof Accept) {
            // TODO outputState -> Game
            throw new CordaRuntimeException("Unimplemented!");
        } else 
        if (command instanceof Reject) {
            requireThat(trx.getInputContractStates().size() == 1, REJECT_INPUT_STATE);
            requireThat(trx.getOutputContractStates().size() == 1, REJECT_OUTPUT_STATE);

            GameProposalState inputState = trx.getInputStates(GameProposalState.class).get(0);
            requireThat(inputState != null, REJECT_INPUT_STATE);
            
            GameProposalResolutionState outputState = trx.getOutputStates(GameProposalResolutionState.class).get(0);
            requireThat(outputState != null, REJECT_OUTPUT_STATE);

            requireThat(outputState.outcome == GameProposalResolutionState.Resolution.REJECT, REJECT_OUTPUT_OUTCOME);
        } else 
        if (command instanceof Cancel) {
            requireThat(trx.getInputContractStates().size() == 1, CANCEL_INPUT_STATE);
            requireThat(trx.getOutputContractStates().size() == 1, CANCEL_OUTPUT_STATE);

            GameProposalState inputState = trx.getInputStates(GameProposalState.class).get(0);
            requireThat(inputState != null, CANCEL_INPUT_STATE);
            
            GameProposalResolutionState outputState = trx.getOutputStates(GameProposalResolutionState.class).get(0);
            requireThat(outputState != null, CANCEL_OUTPUT_STATE);

            requireThat(outputState.outcome == GameProposalResolutionState.Resolution.CANCEL, CANCEL_OUTPUT_OUTCOME);
        } else {
            throw new CordaRuntimeException(UNKNOWN_COMMAND);
        }
    }
    
    private void requireThat(boolean asserted, String errorMessage) {
        if(!asserted) {
            throw new CordaRuntimeException("Failed requirement: " + errorMessage);
        }
    }

    static final String REQUIRE_SINGLE_COMMAND = "Require a single command";
    static final String UNKNOWN_COMMAND = "Unsupported command";

    static final String CREATE_INPUT_STATE  = "Create command should have no input states";
    static final String CREATE_OUTPUT_STATE = "Create command should output exactly one GameProposal state";
    static final String NON_NULL_RECIPIENT_COLOR = "GameProposal.recipientColor field can not be null";

    static final String REJECT_INPUT_STATE  = "Reject command should have exactly one GameProposal input state";
    static final String REJECT_OUTPUT_STATE = "Reject command should have exactly one GameProposalResolution output states";
    static final String REJECT_OUTPUT_OUTCOME = "Reject output state should have Resolution value set to REJECT";

    static final String CANCEL_INPUT_STATE  = "Cancel command should have exactly one GameProposal input state";
    static final String CANCEL_OUTPUT_STATE = "Cancel command should have exactly one GameProposalResolution output states";
    static final String CANCEL_OUTPUT_OUTCOME = "Cancel output state should have Resolution value set to CANCEL";
}
