package djmil.cordacheckers.gameproposal;

import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.stream.Collectors;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import djmil.cordacheckers.FlowResult;
import djmil.cordacheckers.contracts.GameProposalContract;
import djmil.cordacheckers.states.GameProposalResolutionState;
import djmil.cordacheckers.states.GameProposalState;
import net.corda.v5.application.flows.ClientRequestBody;
import net.corda.v5.application.flows.ClientStartableFlow;
import net.corda.v5.application.flows.CordaInject;
import net.corda.v5.application.flows.InitiatingFlow;
import net.corda.v5.application.marshalling.JsonMarshallingService;
import net.corda.v5.application.membership.MemberLookup;
import net.corda.v5.application.messaging.FlowMessaging;
import net.corda.v5.application.messaging.FlowSession;
import net.corda.v5.base.annotations.Suspendable;
import net.corda.v5.base.exceptions.CordaRuntimeException;
import net.corda.v5.base.types.MemberX500Name;
import net.corda.v5.ledger.utxo.Command;
import net.corda.v5.ledger.utxo.StateAndRef;
import net.corda.v5.ledger.utxo.UtxoLedgerService;
import net.corda.v5.ledger.utxo.transaction.UtxoSignedTransaction;
import net.corda.v5.ledger.utxo.transaction.UtxoTransactionBuilder;

import java.time.Duration;
import java.time.Instant;

@InitiatingFlow(protocol = "game-proposal-action")
public class ActionFlow implements ClientStartableFlow {

    private final static Logger log = LoggerFactory.getLogger(CreateFlow.class);

    @CordaInject
    public JsonMarshallingService jsonMarshallingService;

    @CordaInject
    public UtxoLedgerService ledgerService;

    @CordaInject
    public FlowMessaging flowMessaging;

    @CordaInject
    public MemberLookup memberLookup;

    private final static Map<GameProposalResolutionState.Resolution, Command> resoultion2command = Map.ofEntries(
        Map.entry(GameProposalResolutionState.Resolution.CANCEL, new GameProposalContract.Cancel()),
        Map.entry(GameProposalResolutionState.Resolution.REJECT, new GameProposalContract.Reject()),
        Map.entry(GameProposalResolutionState.Resolution.ACCEPT, new GameProposalContract.Accept())
    );

    @Override
    @Suspendable
    public String call(ClientRequestBody requestBody) {
        try {
            ActionFlowArgs args = requestBody.getRequestBodyAs(jsonMarshallingService, ActionFlowArgs.class);

            StateAndRef<GameProposalState> inputState = findUnconsumedGameProposalState(args.getGameProposalUuid());
        
            GameProposalResolutionState outputState = new GameProposalResolutionState(
                args.getAction(),
                inputState.getState().getContractState()
                );

            String trxResult = doTrunsaction(inputState, outputState);

            return new FlowResult(trxResult).toJsonEncodedString(jsonMarshallingService);
        } 
        catch (Exception e) {
            log.warn("GameProposalAction flow failed to process utxo request body " + requestBody + 
                        " because: " + e.getMessage());
            return new FlowResult(e).toJsonEncodedString(jsonMarshallingService);
        }
    }

    @Suspendable
    private StateAndRef<GameProposalState> findUnconsumedGameProposalState (UUID gameProposalUuid) {
        /*
         * Get list of all unconsumed aka 'actuve' GameProposalStates, then filter by UUID.
         * Note, this is an inefficient way to perform this operation if there are a large
         * number of 'active' GameProposals exists in storage.
         */
        List<StateAndRef<GameProposalState>> stateAndRefs = this.ledgerService
            .findUnconsumedStatesByType(GameProposalState.class)
            .stream()
            .filter(sar -> sar.getState().getContractState().getId().equals(gameProposalUuid))
            .collect(Collectors.toList());

        if (stateAndRefs.size() != 1) {
            throw new CordaRuntimeException("Expected only one GameProposal state with id " + gameProposalUuid +
                                            ", but found " + stateAndRefs.size());
        }
        
        return stateAndRefs.get(0);
    }

    @Suspendable
    private String doTrunsaction(StateAndRef<GameProposalState> inputState, GameProposalResolutionState outputState) {
        UtxoTransactionBuilder txBuilder = ledgerService.createTransactionBuilder()
            .setNotary(inputState.getState().getNotaryName())
            .setTimeWindowBetween(Instant.now(), Instant.now().plusMillis(Duration.ofDays(1).toMillis()))
            .addInputState(inputState.getRef())
            .addOutputState(outputState)
            .addCommand(resoultion2command.get(outputState.outcome))
            .addSignatories(outputState.getParticipants());

        UtxoSignedTransaction signedTransaction = txBuilder.toSignedTransaction();

        FlowSession session = flowMessaging.initiateFlow(
            outputState.getRecipient(inputState.getState().getContractState())
            );

        List<FlowSession> sessionsList = Arrays.asList(session);

        ledgerService.finalize(signedTransaction, sessionsList);

        return outputState.getOutcome()+"ED"; // REJECT+ED
    }

}
